#ifdef HAVE_LIBXPM
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/xpm.h>
#include "../include/string.h"
#include "../include/fio.h"
#include "imgio.h"


#ifndef DEBUG_LEVEL
# define DEBUG_LEVEL	0
#endif


/*
 *	RGBA Color:
 */
typedef struct _rgba_color_struct		rgba_color_struct;
struct _rgba_color_struct {
	u_int8_t	r, g, b, a;
};
#define RGBA_COLOR(p)	((rgba_color_struct *)(p))

/*
 *	Symbolic RGBA Color:
 */
typedef struct _sym_rgba_color_struct	sym_rgba_color_struct;
struct _sym_rgba_color_struct {
	char		*name;
	u_int8_t	r, g, b, a;
};
#define SYM_RGBA_COLOR(p)   ((sym_rgba_color_struct *)(p))


/* Last error message pointer */
extern const char *imgio_last_load_error;
extern const char *imgio_last_write_error;


/* Color Utilities */
static void ImgXPMGetSymbolicColor(
	rgba_color_struct *c, const char *name
);
static int ImgXPMColormapColorExists(
	rgba_color_struct **colormap, unsigned int ncolors,
	const u_int8_t *rgba		/* 4 bytes RGBA */
);
static unsigned int ImgXPMColormapMatchColorFromRGBA(
	rgba_color_struct **colormap, unsigned int ncolors,
	const u_int8_t *rgba,		/* 4 bytes RGBA */
	int closeness			/* In bytes, 0 for exact */
);
static rgba_color_struct *ImgXPMColormapColorAppend(
	rgba_color_struct ***colormap, unsigned int *ncolors,
	const u_int8_t *rgba
);


/* XPM Library Version */
void ImgXPMVersion(int *major, int *minor, int *release);


/* XPM Reading */
static char *ImgXPMReadCID(const char *filename);
int ImgXPMReadRGBA(
	const char *filename,
	int *width_rtn, int *height_rtn, int *bpp_rtn, int *bpl_rtn,
	u_int8_t **rgba_rtn,    
	u_int8_t *bg_color,     /* 4 bytes in RGBA format, will be modified */
	int *x_rtn, int *y_rtn,
	int *base_width_rtn, int *base_height_rtn,                  
	char **creator_rtn, char **title_rtn,
	char **author_rtn, char **comments_rtn,
	u_int8_t def_alpha_value,
	void *client_data,      
	int (*progress_cb)(
		void *,         /* Client data */
		int, int,       /* Progress current & end values */
		int, int,       /* Width & height */
		int, int,       /* Bytes per line & bytes per pixel */
		u_int8_t *      /* RGBA data */
	),
	int *user_aborted
);


/* XPM Writing */
static int ImgXPMWriteFileRGBAToColor(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments,
	int max_colors,		/* -1 for no color limit */
	u_int8_t threshold,
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
);
static int ImgXPMWriteFileRGBAToGreyscale(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments,
	int max_colors,		/* -1 for no color limit */
	u_int8_t threshold,
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */ 
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
);
static int ImgXPMWriteFileRGBAToBW(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments, 
	int max_colors,		/* -1 for no color limit */
	u_int8_t threshold,
	void *client_data,
	int (*progress_cb)(       
		void *,		/* Client data */
		int, int,	/* Progress current & end values */ 
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
);
int ImgXPMWriteFileRGBA(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments,
	int format,		/* 0 = Black & White
				 * 1 = Greyscale    
				 * 2 = Color */ 
	int max_colors,		/* -1 for no color limit */
	u_int8_t threshold,
	void *client_data,  
	int (*progress_cb)( 
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	)
);


/*
 *	XPM Pixel ASCII Characters:
 *
 *	These are ASCII character values that represent a range of
 *	ASCII characters that may be used as characters to represent
 *	a pixel value in an XPM file.
 *
 *	The range here must not contain the '"' or '\' characters since
 *	these characters will be parsed by C interpreters when the
 *	XPM file is used in C source code.
 */
#define XPM_ASCII_CHAR_MIN		35	/* '#' */
#define XPM_ASCII_CHAR_MAX		90	/* 'Z' */
#define XPM_ASCII_CHAR_RANGE		(XPM_ASCII_CHAR_MAX - XPM_ASCII_CHAR_MIN + 1)


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)

/*
 *	Symbolic Colors List:
 */
#define IMG_XPM_SYM_RGBA_COLOR_LIST	{		\
{	"Black",	0x00, 0x00, 0x00, 0xff },	\
{	"White",	0xff, 0xff, 0xff, 0xff },	\
{	"None",		0x00, 0x00, 0x00, 0x00 },	\
{	"Transparent",	0x00, 0x00, 0x00, 0x00 },	\
{	"Background",	0x00, 0x00, 0x00, 0x00 },	\
{	"LightGrey",	0xc0, 0xc0, 0xc0, 0xff },	\
{	"Grey",		0x80, 0x80, 0x80, 0xff },	\
{	"DarkGrey",	0x40, 0x40, 0x40, 0xff },	\
{	"Red",		0xff, 0x00, 0x00, 0xff },	\
{	"Green",	0x00, 0xff, 0x00, 0xff },	\
{	"Blue",		0x00, 0x00, 0xff, 0xff }	\
}
static sym_rgba_color_struct sym_rgba_color_list[] = IMG_XPM_SYM_RGBA_COLOR_LIST;


/*
 *	Sets the RGBA color to the symbolic color name specified by
 *	name.
 */
static void ImgXPMGetSymbolicColor(
	rgba_color_struct *c, const char *name
)
{
	const sym_rgba_color_struct *sym_c;
	int i, m = sizeof(sym_rgba_color_list) / sizeof(sym_rgba_color_struct);

	for(i = 0; i < m; i++)
	{
	    sym_c = &sym_rgba_color_list[i];
	    if(!strcasecmp(name, sym_c->name))
	    {
#if (DEBUG_LEVEL >= 2)
printf(
 "ImgXPMGetSymbolicColor():\
 Matched symbolic color \"%s\" to RGBA=%.2x%.2x%.2x%.2x\n",
 sym_c->name,
 sym_c->r,
 sym_c->g,
 sym_c->b,
 sym_c->a
);
#endif
		c->r = sym_c->r;
		c->g = sym_c->g;
		c->b = sym_c->b;
		c->a = sym_c->a;
		break;
	    }
	}	    
}

/*
 *	Checks if the RGBA color exists in the RGBA colormap.
 */
static int ImgXPMColormapColorExists(
	rgba_color_struct **colormap, unsigned int ncolors,
	const u_int8_t *rgba		/* 4 bytes RGBA */
)
{
	return((ImgXPMColormapMatchColorFromRGBA(
	    colormap, ncolors, rgba, 0) == (unsigned int)-1) ? 0 : 1
	);
}

/*
 *	Returns a color index of a color on the RGBA colormap that
 *	matches the specified RGBA color with the specified closeness.
 *
 *	Inputs assumed valid.
 *
 *	Can return (unsigned int)-1 on failed match.
 */
static unsigned int ImgXPMColormapMatchColorFromRGBA(
	rgba_color_struct **colormap, unsigned int ncolors,
	const u_int8_t *rgba,		/* 4 bytes RGBA */
	int closeness			/* In bytes, 0 for exact */
)
{
	unsigned int i;
	const rgba_color_struct *c;

	if(closeness == 0)
	{
	    for(i = 0; i < ncolors; i++)
	    {
		c = colormap[i];
		if((c->r == rgba[0]) &&
		   (c->g == rgba[1]) &&
		   (c->b == rgba[2]) &&
		   (c->a == rgba[3])
		)
		    return(i);
	    }
	}
	else
	{
	    for(i = 0; i < ncolors; i++)
	    {
		c = colormap[i];
		if((((int)c->r + closeness) >= rgba[0]) &&
		   (((int)c->r - closeness) <= rgba[0]) &&
		   (((int)c->g + closeness) >= rgba[1]) &&
		   (((int)c->g - closeness) <= rgba[1]) &&
		   (((int)c->b + closeness) >= rgba[2]) &&
		   (((int)c->b - closeness) <= rgba[2]) &&
		   (((int)c->a + closeness) >= rgba[3]) &&
		   (((int)c->a - closeness) <= rgba[3])
		)
		    return(i);
	    }

	}
	return((unsigned int)-1);
}

/*
 *	Appends a new color to the RGBA colormap.
 */
static rgba_color_struct *ImgXPMColormapColorAppend(
	rgba_color_struct ***colormap, unsigned int *ncolors,
	const u_int8_t *rgba
)
{
	rgba_color_struct *c;
	unsigned int i = MAX(*ncolors, 0);
	*ncolors = i + 1;
	*colormap = (rgba_color_struct **)realloc(
	    *colormap,
	    (*ncolors) * sizeof(rgba_color_struct *)
	);
	if(*colormap == NULL)
	{
	    *ncolors = 0;
	    return(NULL);
	}

	(*colormap)[i] = c = RGBA_COLOR(
	    malloc(sizeof(rgba_color_struct))
	);
	if(c == NULL)
	{
	    *ncolors = (*ncolors) - 1;
	    return(NULL);
	}

	c->r = rgba[0];
	c->g = rgba[1];
	c->b = rgba[2];
	c->a = rgba[3];

#if (DEBUG_LEVEL >= 3)
printf(
 "ImgXPMColormapColorAppend(): Appended color %i = 0x%.2x%.2x%.2x%.2x\n",
 i, c->r, c->g, c->b, c->a
);
#endif

	return(c);
}


/*
 *	Gets the C ID from the XPM file.
 */
static char *ImgXPMReadCID(const char *filename)
{
	const int buf_len = 4 * 80;	/* First 4 lines */
	char *buf, *s, *c_id = NULL;
	int bytes_read;
	FILE *fp = FOpen(filename, "rb");
	if(fp == NULL)
	    return(NULL);

	buf = (char *)malloc((buf_len + 1) * sizeof(char));
	if(buf == NULL)
	{
	    FClose(fp);
	    return(NULL);
	}

	/* Read a few characters at the start of the file, hoping to
	 * find the C ID
	 */
	bytes_read = fread(buf, sizeof(char), buf_len, fp);

	FClose(fp);

	if((bytes_read > 0) && (bytes_read <= buf_len))
	{
	    buf[bytes_read] = '\0';

	    /* Seek to type declaration of C ID (if any) */
	    s = strstr(buf, "char");
	    if(s != NULL)
		s = strchr(s, '*');
	    if(s != NULL)
	    {
		/* Found declaration of C ID, now see s to start of the
		 * C ID
		 */
		char *s_end;

		while(*s == '*')
		    s++;
		while(ISSPACE(*s))
		    s++;

		/* Seek s_end to end of C ID */
		s_end = s;
		while((*s_end != '\0') && (*s_end != '[') &&
		      (*s_end != '=') && !ISSPACE(*s_end)
		)
		    s_end++;

		/* Copy C ID to return string */
		if(s_end > s)
		{
		    int c_id_len = s_end - s;
		    c_id = (char *)malloc((c_id_len + 1) * sizeof(char));
		    if(c_id != NULL)
		    {
			memcpy(c_id, s, c_id_len);
			c_id[c_id_len] = '\0';
		    }
		}
	    }
	}

	free(buf);

	return(c_id);
}


/*
 *	Gets the XPM library's version.
 */
void ImgXPMVersion(int *major, int *minor, int *release)
{
	if(major != NULL)
	    *major = XpmFormat;
	if(minor != NULL)
	    *minor = XpmVersion;
	if(release != NULL)
	    *release = XpmRevision;
}


/*
 *	Reads the XPM file.
 *
 *      Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value (invalid format/not a xpm file)
 *      -3      Systems error
 *      -4      User abort
 */
int ImgXPMReadRGBA(       
	const char *filename,
	int *width_rtn, int *height_rtn, int *bpp_rtn, int *bpl_rtn,
	u_int8_t **rgba_rtn,    
	u_int8_t *bg_color,     /* 4 bytes in RGBA format, will be modified */
	int *x_rtn, int *y_rtn,
	int *base_width_rtn, int *base_height_rtn,
	char **creator_rtn, char **title_rtn,
	char **author_rtn, char **comments_rtn,
	u_int8_t def_alpha_value,
	void *client_data,      
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
)
{
	const int bpp = 4;
	int width, height, bpl;
	unsigned long info_flags;
	XpmInfo *info = NULL;
	XpmImage *img = NULL;
	u_int8_t *rgba = NULL;
	unsigned int total_rgba_colors = 0;
	rgba_color_struct **rgba_color = NULL;

#define FREE_ALL	{		\
 unsigned int i;			\
					\
 /* Delete XPM image */			\
 if(img != NULL) {			\
  XpmFreeXpmImage(img);			\
  free(img);				\
  img = NULL;				\
 }					\
					\
 /* Delete XPM info */			\
 if(info != NULL) {			\
  XpmFreeXpmInfo(info);			\
  free(info);				\
  info = NULL;				\
 }					\
					\
 /* Delete RGBA image data */		\
 free(rgba);				\
 rgba = NULL;				\
					\
 /* Delete RGBA colormap */		\
 for(i = 0; i < total_rgba_colors; i++)	{ \
  free(rgba_color[i]);			\
 }					\
 free(rgba_color);			\
 rgba_color = NULL;			\
 total_rgba_colors = 0;			\
					\
}

	/* Reset last load error */
	imgio_last_load_error = NULL;

	/* Skip if user aborted */
	if(*user_aborted)
	{
	    imgio_last_load_error = "User aborted operation";
	    return(-4);
	}

	/* Allocate data */
	img = (XpmImage *)calloc(1, sizeof(XpmImage));
	info = (XpmInfo *)calloc(1, sizeof(XpmInfo));	

	/* Read XPM file and get XpmImage and XpmInfo */
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMReadRGBA(): Reading XPM file \"%s\"...\n",
 filename
);       
#endif   
	switch(XpmReadFileToXpmImage((char *)filename, img, info))
	{
	  case XpmColorError:
	    FREE_ALL
	    imgio_last_load_error = "XPM color error"; 
	    return(-1);
	    break;
	  case XpmSuccess:
	    break;
	  case XpmOpenFailed:
	    FREE_ALL 
	    imgio_last_load_error = "Unable to open XPM file for reading"; 
	    return(-1);
	    break;
	  case XpmFileInvalid:
	    FREE_ALL
	    imgio_last_load_error = "Invalid XPM file";
	    return(-2);
	    break;
	  case XpmNoMemory:
	    FREE_ALL
	    imgio_last_load_error = "Memory allocation error"; 
	    return(-3);
	    break;
	  case XpmColorFailed:
	    FREE_ALL
	    imgio_last_load_error = "XPM color failed";
	    return(-1);
	    break;
	  default:
	    FREE_ALL
	    imgio_last_load_error = "Error reading XPM file";
	    return(-1);
	    break;
	}

	/* Get XPM values */
	info_flags = info->valuemask;
	width = img->width;
	height = img->height;
	bpl = width * bpp;

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMReadRGBA(): size: %ix%i  hotspot: %i %i  colors: %i  chars/pixel: %i\n",
 width, height,
 (info_flags & XpmHotspot) ? (int)info->x_hotspot : 0,
 (info_flags & XpmHotspot) ? (int)info->y_hotspot : 0,
 img->ncolors, img->cpp
);
if(info_flags & XpmComments) {
 if(info->hints_cmt != NULL)
  printf(
   "ImgXPMReadRGBA(): Hints Comment: \"%s\"\n",
   info->hints_cmt
  );
 if(info->colors_cmt != NULL)
  printf(
   "ImgXPMReadRGBA(): Colors Comment: \"%s\"\n",
   info->colors_cmt
  );
 if(info->pixels_cmt != NULL)
  printf(
   "ImgXPMReadRGBA(): Image Comment: \"%s\"\n",
   info->pixels_cmt
  );
}
#endif

	if((width <= 0) || (height <= 0))
	{
	    FREE_ALL
	    imgio_last_load_error = "Invalid image size";
	    return(-2);
	}

	if(img->data == NULL)
	{
	    FREE_ALL
	    imgio_last_load_error = "Memory allocation error";
	    return(-3);
	}              

	/* Create RGBA colormap from the XPM colormap */
	total_rgba_colors = img->ncolors;
	if(total_rgba_colors > 0)
	{
	    const char *s;
	    unsigned int i;
	    rgba_color_struct *c;
	    const XpmColor *xpm_color = img->colorTable, *xpm_c;

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMReadRGBA(): Creating RGBA colormap with %i colors from XPM colormap.\n",
 total_rgba_colors
);
#endif

	    /* Allocate RGBA colormap */
	    rgba_color = (rgba_color_struct **)calloc(
		total_rgba_colors, sizeof(rgba_color_struct *)
	    );
	    if(rgba_color == NULL)
	    {
		total_rgba_colors = 0;
		FREE_ALL  
		imgio_last_load_error = "Memory allocation error";
		return(-3);
	    }

	    /* Allocate and set each RGBA color */
	    for(i = 0; i < total_rgba_colors; i++)
	    {
		xpm_c = &xpm_color[i];
		rgba_color[i] = c = RGBA_COLOR(
		    malloc(sizeof(rgba_color_struct))
		);
		if(c == NULL)
		{
		    FREE_ALL  
		    imgio_last_load_error = "Memory allocation error";
		    return(-3);
		}

		/* Symbolic color? */
		s = xpm_c->symbolic;
		if(s != NULL)
		{
		    ImgXPMGetSymbolicColor(c, s);
		    continue;
		}

		/* Get color string and parse it */
		s = xpm_c->c_color;
		if(s != NULL)
		{
		    /* Symbolic color? */
		    if(*s != '#')
		    {
			ImgXPMGetSymbolicColor(c, s);
			continue;
		    }

#if (DEBUG_LEVEL >= 3)
printf(
 "ImgXPMReadRGBA(): Color %i=\"%s\"\n",
 i, s
);
#endif

		    /* Seek past blanks and '#' prefix */
		    while((*s == ' ') || (*s == '\t') || (*s == '#'))
			s++;
		    if(STRLEN(s) >= 6)
		    {
#define HEX_CTOI(_c_)   (u_int8_t)(((_c_) >= 'a') ?     \
 ((_c_) - 'a' + 10) : ((_c_) - '0')                     \
)
			c->r = (HEX_CTOI(tolower(s[0])) * 16) +
			    HEX_CTOI(tolower(s[1]));
			c->g = (HEX_CTOI(tolower(s[2])) * 16) +
			    HEX_CTOI(tolower(s[3]));
			c->b = (HEX_CTOI(tolower(s[4])) * 16) +
			    HEX_CTOI(tolower(s[5]));
			c->a = 0xff;
#undef HEX_CTOI
		    }
		    continue;
		}

	    }
	}

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMReadRGBA():\
 Creating RGBA image data from XPM color-index image data...\n"
);
#endif

	/* Allocate RGBA return image data */
	rgba = (u_int8_t *)malloc(bpl * height);
	if(rgba == NULL)
	{
	    FREE_ALL
	    imgio_last_load_error = "Memory allocation error";
	    return(-3);
	}

	/* Report initial progress */
	if((progress_cb != NULL) && !(*user_aborted))
	{
	    if(!progress_cb(
		client_data,
		0, height,
		width, height,
		bpl, bpp,
		rgba
	    ))
		*user_aborted = 1;
	}

	/* Copy/convert the loaded XPM image data to the return RGBA
	 * image data
	 */
	if(!(*user_aborted))
	{
	    int y = 0;
	    unsigned int i;
	    const unsigned int	*src_line = img->data,
				*src_line_end = src_line + (width * height),
				*src_ptr, *src_end;
	    u_int8_t	*tar_line = rgba,
			*tar_ptr;
	    const rgba_color_struct *c;

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMReadRGBA():\
 Converting XPM image data to RGBA image data...\n"
);
#endif

	    /* Iterate through source XPM image data */
	    while(src_line < src_line_end)
	    {
		src_ptr = src_line;
		src_end = src_ptr + width;
		tar_ptr = tar_line;

		/* Iterate through current line on the source XPM image
		 * data
		 */
		while(src_ptr < src_end)
		{
		    /* Get color index value from the current source
		     * XPM image data pixel value
		     */
		    i = *src_ptr++;

		    /* Get corresponding color from the RGBA colormap */
		    c = (i < total_rgba_colors) ?
			c = rgba_color[i] : NULL;

		    /* Set target RGBA image data pixel if a color was
		     * matched
		     */
		    if(c != NULL)
		    {
			*tar_ptr++ = c->r;
			*tar_ptr++ = c->g;
			*tar_ptr++ = c->b;
			*tar_ptr++ = c->a;
		    }
		    else
		    {
			tar_ptr += bpp;
		    }
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			y, height,
			width, height,
			bpl, bpp,
			rgba
		    ))
		    {
			*user_aborted = 1;
			break;
		    }
		}

		y++;
		src_line += width;
		tar_line += bpl;
	    }
	}


	/* Handle extensions */
	if(info_flags & XpmExtensions)
	{
	    unsigned int i;
	    XpmExtension *ext;

	    for(i = 0; i < info->nextensions; i++)
	    {
		ext = &info->extensions[i];
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMReadRGBA(): Extension \"%s\" encountered.\n",
 ext->name
);
#endif

	    }
	}


	/* Update returns */
	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;
	if(bpl_rtn != NULL)
	    *bpl_rtn = bpl;
	if(bpp_rtn != NULL)
	    *bpp_rtn = bpp;
	if(rgba_rtn != NULL)
	{
	    free(*rgba_rtn);
	    *rgba_rtn = rgba;
	    rgba = NULL;
	}
	if(x_rtn != NULL)
	    *x_rtn = (info_flags & XpmHotspot) ? (int)info->x_hotspot : 0;
	if(y_rtn != NULL)
	    *y_rtn = (info_flags & XpmHotspot) ? (int)info->y_hotspot : 0;
	if(base_width_rtn != NULL)
	    *base_width_rtn = width;
	if(base_height_rtn != NULL)    
	    *base_height_rtn = height;

	/* Get C ID (if any) and return it as the title */
	if(title_rtn != NULL)
	{
	    free(*title_rtn);
	    *title_rtn = ImgXPMReadCID(filename);
	}

	/* Report final progress */
	if((progress_cb != NULL) && !(*user_aborted))
	{
	    if(!progress_cb(                       
		client_data,
		height, height,
		width, height,
		bpl, bpp,
		rgba
	    ))
		*user_aborted = 1;
	}

	FREE_ALL

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMReadRGBA(): Done.\n"
);
#endif

	if(*user_aborted)
	{
	    imgio_last_load_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(0);

#undef FREE_ALL
}


/*
 *	Writes the XpmImage and XpmInfo to file.
 */
static int ImgXPMWriteFileXPMImage(
	const char *filename, const char *c_id,
	XpmImage *img, XpmInfo *info
)
{
	char *xpm_buf, *s;
	FILE *fp;

	/* Generate data from XpmImage and XpmInfo */
	switch(XpmCreateBufferFromXpmImage(&xpm_buf, img, info))
	{
	  case XpmColorError:
	    imgio_last_load_error = "XPM color error";
	    return(-1);
	    break;
	  case XpmSuccess:
	    break;
	  case XpmOpenFailed:
	    imgio_last_load_error = "Unable to open XPM file for writing";
	    return(-1);
	    break;
	  case XpmFileInvalid:
	    imgio_last_load_error = "Invalid XPM file";
	    return(-2);
	    break;
	  case XpmNoMemory:
	    imgio_last_load_error = "Memory allocation error";
	    return(-3);
	    break;
	  case XpmColorFailed:
	    imgio_last_load_error = "XPM color failed";
	    return(-1);
	    break;
	  default:
	    imgio_last_load_error = "Error writing XPM file";
	    return(-1);
	    break;
	}
	if(xpm_buf == NULL)
	{
	    imgio_last_write_error = "Unable to generate XPM buffer";
	    return(-1);
	}

	/* Seek to C ID */
	s = strstr(xpm_buf, "char");
	if(s != NULL)
	    s = strchr(s, '*');
	if(s != NULL)
	{
	    char *s_end;

	    while(*s == '*')
		s++;
	    while(ISSPACE(*s))
		s++;

	    /* Seek s_end to end of C ID */
	    s_end = s;    
	    while((*s_end != '\0') && (*s_end != '[') &&
	          (*s_end != '=') && !ISSPACE(*s_end)
	    )
		s_end++;
	    if(s_end != NULL)
	    {
		/* Delete existing C ID in the XPM buffer and insert
		 * our C ID
		 */
		xpm_buf = strdelchrs(
		    xpm_buf, s - xpm_buf, s_end - s
		);
		xpm_buf = strinsstr(
		    xpm_buf, s - xpm_buf, c_id
		);
	    }
	}

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileXPMImage():\
 Writing XpmImage and XpmInfo to XPM file...\n"
);
#endif

	/* Open XPM file for writing */
	fp = FOpen(filename, "wb");
	if(fp == NULL)
	{
	    if(xpm_buf != NULL)
		XpmFree(xpm_buf);
	    imgio_last_write_error = "Unable to open XPM file for writing";
	    return(-1);  
	}

	/* Write the XPM buffer to the XPM file */
	if(fwrite(xpm_buf, sizeof(char), STRLEN(xpm_buf), fp) <= 0)
	{
	    if(xpm_buf != NULL)
		XpmFree(xpm_buf);
	    FClose(fp);
	    imgio_last_write_error = "Unable to write XPM buffer to the XPM file";
	    return(-1);  
	}

	/* Close XPM file */
	FClose(fp);

	XpmFree(xpm_buf);

	return(0);
}

/*
 *      Writes the specified image data to an XPM file in color.
 */
static int ImgXPMWriteFileRGBAToColor(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments,
	int max_colors,		/* -1 for no color limit */
	u_int8_t threshold,
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
)
{
	unsigned int total_rgba_colors = 0;
	rgba_color_struct **rgba_color = NULL;
	XpmInfo *info = NULL;
	XpmImage *img = NULL;

#define FREE_ALL	{		\
 unsigned int i;			\
					\
 /* Delete XPM image */			\
 if(img != NULL) {			\
  XpmFreeXpmImage(img);			\
  free(img);				\
  img = NULL;				\
 }					\
					\
 /* Delete XPM info */			\
 if(info != NULL) {			\
  XpmFreeXpmInfo(info);			\
  free(info);				\
  info = NULL;				\
 }					\
					\
 /* Delete RGBA colormap */		\
 for(i = 0; i < total_rgba_colors; i++) { \
  free(rgba_color[i]);			\
 }					\
 free(rgba_color);			\
 rgba_color = NULL;			\
 total_rgba_colors = 0;			\
					\
}

	/* Begin generating colormap */

	/* Generate symbolic colors */
	if(1)
	{
	    u_int8_t pix[4] = { 0x00, 0x00, 0x00, 0x00 };
	    ImgXPMColormapColorAppend(
		&rgba_color, &total_rgba_colors,
		pix
	    );
	}


	/* Generate colors from the source RGBA image */
/*	if(max_colors > 0) */
	if(0)
	{
	    /* Generate a limited number of colors */
/* TODO */
	}
	else
	{
	    /* Generate as many colors as needed */
	    int y = 0;
	    const u_int8_t	*line = rgba,
				*line_end = line + (bpl * height),
				*ptr, *ptr_end;

	    /* Iterate through the source RGBA image and generate the
	     * RGBA colormap
	     */
	    while(line < line_end)
	    {
		for(ptr = line,
		    ptr_end = ptr + (width * bpp);
		    ptr < ptr_end;
		    ptr += bpp
		)
		{
		    /* Do not generate colors from a source pixel who's
		     * alpha value is below the threshold
		     */
		    if(ptr[3] < threshold)
			continue;

		    /* Does this source pixel not have a color in the
		     * RGBA colormap yet?
		     */
		    if(!ImgXPMColormapColorExists(
			rgba_color, total_rgba_colors, ptr
		    ))
		    {
			if(ImgXPMColormapColorAppend(
			    &rgba_color, &total_rgba_colors,
			    ptr
			) == NULL)
			{
			    FREE_ALL
			    imgio_last_write_error = "Memory allocation error";
			    return(-3);
			}
		    }
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			y,
			height,
			width, height,
			bpl, bpp,
			rgba
		    ))
		    { 
			*user_aborted = 1;
			break;
		    }
		}

		y++;
		line += bpl;
	    }
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToColor(): Generated colormap with %i colors.\n",
 total_rgba_colors
);
#endif


	/* Create XpmInfo */
	info = (XpmInfo *)calloc(1, sizeof(XpmInfo));
	if(info == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}
	info->valuemask = XpmComments | XpmHotspot;
	info->hints_cmt = NULL;
	info->colors_cmt = NULL;
	info->pixels_cmt = STRDUP(comments);
	info->x_hotspot = x;
	info->y_hotspot = y;


	/* Create XpmImage */
	img = (XpmImage *)calloc(1, sizeof(XpmImage));
	if(img == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error"; 
	    return(-3);
	}
	img->width = width;
	img->height = height;
	if(total_rgba_colors > 0)
	{
	    unsigned int m = XPM_ASCII_CHAR_RANGE;
	    for(img->cpp = 1;
		m < total_rgba_colors;
		img->cpp++, m *= XPM_ASCII_CHAR_RANGE
	    );
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToColor():\
 Generating XPM colormap: ncolors: %i  cpp: %i (%i possible colors)\n",
 total_rgba_colors, img->cpp, m
);
#endif
	}

	/* Create XPM colormap on the XpmImage */
	if((img->cpp > 0) && (total_rgba_colors > 0))
	{
	    const unsigned int cpp = img->cpp;
	    unsigned int i;
	    char *s, *c_string = malloc((cpp + 1) * sizeof(char));
	    const rgba_color_struct *src_c;
	    XpmColor *tar_c;

	    /* Reset color string */
	    memset(c_string, XPM_ASCII_CHAR_MIN, cpp);
	    c_string[cpp] = '\0';

	    /* Allocate XpmColors on the XPM colormap */
	    img->ncolors = total_rgba_colors;
	    img->colorTable = (XpmColor *)calloc(
		img->ncolors, sizeof(XpmColor)
	    );
	    if(img->colorTable == NULL)
	    {
		free(c_string);
		img->ncolors = 0;
		FREE_ALL 
		imgio_last_write_error = "Memory allocation error";
		return(-3);
	    }

	    /* Iterate through RGBA colormap and create XpmColors */
	    for(i = 0; i < total_rgba_colors; i++)
	    {
		src_c = rgba_color[i];
		tar_c = &img->colorTable[i];

		/* Set XpmColor's ASCII characters string that represent
		 * this color
		 */
		tar_c->string = STRDUP(c_string);

		/* Set XpmColor's color value string */
/* TODO check for symbolic colors */
		if(src_c->a == 0x00)
		{
		    tar_c->c_color = STRDUP("None");
		}
		else
		{
		    tar_c->c_color = s = (char *)malloc((7 + 1) * sizeof(char));
		    sprintf(
			s, "#%.2X%.2X%.2X",
			src_c->r, src_c->g, src_c->b
		    );
		}

		/* Increment characters in the color string */
		for(s = c_string; *s != '\0'; s++)
		{
		    if(*s < XPM_ASCII_CHAR_MAX)
		    {
			*s = (*s) + 1;
			break;
		    }
		    else
		    {
			*s = XPM_ASCII_CHAR_MIN;
		    }
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			i,
			total_rgba_colors + height,
			width, height,
			bpl, bpp,
			rgba
		    ))
		    { 
			*user_aborted = 1;
			break;
		    }
		}

#if (DEBUG_LEVEL >= 3)
printf(
 "ImgXPMWriteFileRGBAToColor():\
 Generated color %i: \"%s\"=\"%s\"\n",
 i, tar_c->string, tar_c->c_color
);
#endif

	    }

	    free(c_string);
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}


	/* Allocate XPM image data */
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToColor():\
 Generating XpmImage data %ix%i  %i bytes...\n",
 width, height,
 width * height * sizeof(unsigned int)
);
#endif
	img->data = (unsigned int *)malloc(
	    width * height * sizeof(unsigned int)
	);
	if(img->data == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}

	/* Copy/convert the source RGBA image data to the target XPM
	 * color-index image data
	 */
	if(img->data != NULL)
	{
	    u_int32_t transparent_rgba = 0x00000000;
	    unsigned int xpm_transparent_pix = ImgXPMColormapMatchColorFromRGBA(
		rgba_color, total_rgba_colors,
		(const u_int8_t *)&transparent_rgba, 0
	    );
	    unsigned int *tar_ptr = img->data;
	    int y = 0;
	    const u_int8_t	*src_line = rgba,
				*src_line_end = src_line + (bpl * height),
				*src_ptr, *src_end;
	    while(src_line < src_line_end)
	    {
		src_ptr = src_line;
		src_end = src_ptr + (width * bpp);
		while(src_ptr < src_end)
		{
		    if(src_ptr[3] < threshold)
			*tar_ptr++ = xpm_transparent_pix;
		    else
			*tar_ptr++ = ImgXPMColormapMatchColorFromRGBA(
			    rgba_color, total_rgba_colors,
			    src_ptr,
			    0		/* Closeness, 0 for exact */
			);
		    src_ptr += bpp;
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			total_rgba_colors + y,
			total_rgba_colors + height,
			width, height,
			bpl, bpp,
			rgba
		    ))
		    {
			*user_aborted = 1;
			break;
		    }
		}		

		y++;
		src_line += bpl;
	    }
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}


	/* Write XpmImage and XpmInfo to the XPM file */
	if(ImgXPMWriteFileXPMImage(filename, c_id, img, info))
	{
	    FREE_ALL
	    return(-1);
	}
	else
	{
	    FREE_ALL
	    return(0);
	}
#undef FREE_ALL
}

/*
 *	Writes the specified image data to an XPM file in greyscale
 */
static int ImgXPMWriteFileRGBAToGreyscale(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments,
	int max_colors,         /* -1 for no color limit */                     
	u_int8_t threshold,
	void *client_data,
	int (*progress_cb)(       
		void *,         /* Client data */
		int, int,       /* Progress current & end values */ 
		int, int,       /* Width & height */
		int, int,       /* Bytes per line & bytes per pixel */
		u_int8_t *      /* RGBA data */
	),
	int *user_aborted
)
{
	unsigned int total_rgba_colors = 0;
	rgba_color_struct **rgba_color = NULL;
	XpmInfo *info = NULL;
	XpmImage *img = NULL;

#define FREE_ALL	{		\
 unsigned int i;			\
					\
 /* Delete XPM image */			\
 if(img != NULL) {			\
  XpmFreeXpmImage(img);			\
  free(img);				\
  img = NULL;				\
 }					\
					\
 /* Delete XPM info */			\
 if(info != NULL) {			\
  XpmFreeXpmInfo(info);			\
  free(info);				\
  info = NULL;				\
 }					\
					\
 /* Delete RGBA colormap */		\
 for(i = 0; i < total_rgba_colors; i++) { \
  free(rgba_color[i]);			\
 }					\
 free(rgba_color);			\
 rgba_color = NULL;			\
 total_rgba_colors = 0;			\
					\
}
 
	/* Begin generating colormap */  

	/* Generate symbolic colors */   
	if(1)
	{
	    u_int8_t pix[4] = { 0x00, 0x00, 0x00, 0x00 };
	    ImgXPMColormapColorAppend(   
		&rgba_color, &total_rgba_colors,
		pix
	    );
	}

	/* Generate greyscale RGBA colormap from the source RGBA image */
	if(1)
	{
	    int y = 0;
	    u_int32_t g, grey_rgba;
	    const u_int8_t	*line = rgba,
				*line_end = line + (bpl * height),
				*ptr, *ptr_end;

	    /* Iterate through the source RGBA image and generate the
	     * RGBA colormap
	     */
	    while(line < line_end)
	    {
		for(ptr = line,
		    ptr_end = ptr + (width * bpp);
		    ptr < ptr_end;
		    ptr += bpp
		)
		{
		    /* Do not generate colors from a source pixel who's
		     * alpha value is below the threshold since those
		     * pixels will always use the transparent color
		     */
		    if(ptr[3] < threshold)
			continue;

		    /* Get grey value of the current RGBA pixel */
		    g = ((u_int32_t)ptr[0] + (u_int32_t)ptr[1] +
			(u_int32_t)ptr[2]) / 3;
		    grey_rgba = (u_int32_t)(
			0xff000000 |
			(g << 16) | (g << 8) | (g << 0)
		    );

		    /* Is this grey pixel value not in the RGBA
		     * colormap yet?
		     */
		    if(!ImgXPMColormapColorExists(
			rgba_color, total_rgba_colors,
			(const u_int8_t *)&grey_rgba
		    ))
		    { 
			if(ImgXPMColormapColorAppend(
			    &rgba_color, &total_rgba_colors,
			    (const u_int8_t *)&grey_rgba
			) == NULL)
			{
			    FREE_ALL
			    imgio_last_write_error = "Memory allocation error";
			    return(-3);
			}
		    }
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			y,
			height,
			width, height,
			bpl, bpp,
			rgba
		    )) 
		    {
			*user_aborted = 1;
			break;
		    }
		}

		y++;
		line += bpl;
	    }
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToGreyscale(): Generated colormap with %i colors.\n",
 total_rgba_colors
);
#endif


	/* Create XpmInfo */
	info = (XpmInfo *)calloc(1, sizeof(XpmInfo));
	if(info == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}
	info->valuemask = XpmComments | XpmHotspot;
	info->hints_cmt = NULL;
	info->colors_cmt = NULL; 
	info->pixels_cmt = STRDUP(comments);
	info->x_hotspot = x;
	info->y_hotspot = y;


	/* Create XpmImage */
	img = (XpmImage *)calloc(1, sizeof(XpmImage));
	if(img == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}
	img->width = width;
	img->height = height;
	if(total_rgba_colors > 0)       
	{                       
	    unsigned int m = XPM_ASCII_CHAR_RANGE;
	    for(img->cpp = 1;
		m < total_rgba_colors;
		img->cpp++, m *= XPM_ASCII_CHAR_RANGE
	    );
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToColor():\
 Generating XPM colormap: ncolors: %i  cpp: %i (%i possible colors)\n",
 total_rgba_colors, img->cpp, m
);
#endif
	}

	/* Create XPM colormap on the XpmImage */
	if((img->cpp > 0) && (total_rgba_colors > 0))
	{
	    const unsigned int cpp = img->cpp;
	    unsigned int i;
	    char *s, *c_string = malloc((cpp + 1) * sizeof(char));
	    const rgba_color_struct *src_c;
	    XpmColor *tar_c;

	    /* Reset color string */
	    memset(c_string, XPM_ASCII_CHAR_MIN, cpp);
	    c_string[cpp] = '\0';       

	    /* Allocate XpmColors on the XPM colormap */
	    img->ncolors = total_rgba_colors;
	    img->colorTable = (XpmColor *)calloc(
		img->ncolors, sizeof(XpmColor)
	    );
	    if(img->colorTable == NULL)
	    {
		free(c_string);
		img->ncolors = 0;
		FREE_ALL
		imgio_last_write_error = "Memory allocation error";
		return(-3);
	    }

	    /* Iterate through RGBA colormap and create XpmColors */
	    for(i = 0; i < total_rgba_colors; i++)
	    {
		src_c = rgba_color[i];
		tar_c = &img->colorTable[i];

		/* Set XpmColor's ASCII characters string that represent
		 * this color
		 */
		tar_c->string = STRDUP(c_string);       

		/* Set XpmColor's color value string */
/* TODO check for symbolic colors */
		if(src_c->a == 0x00)
		{
		    tar_c->c_color = STRDUP("None");
		}
		else
		{
		    tar_c->c_color = s = (char *)malloc((7 + 1) * sizeof(char));
		    sprintf(
			s, "#%.2X%.2X%.2X",
			src_c->r, src_c->g, src_c->b
		    );
		}

		/* Increment characters in the color string */
		for(s = c_string; *s != '\0'; s++)
		{
		    if(*s < XPM_ASCII_CHAR_MAX)
		    {
			*s = (*s) + 1;
			break;
		    }
		    else
		    {
			*s = XPM_ASCII_CHAR_MIN;
		    }
		}    

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			i,
			total_rgba_colors + height, 
			width, height,
			bpl, bpp,
			rgba
		    ))
		    {
			*user_aborted = 1;
			break;
		    }
		}

#if (DEBUG_LEVEL >= 3)
printf(
 "ImgXPMWriteFileRGBAToGreyscale():\
 Generated color %i: \"%s\"=\"%s\"\n",
 i, tar_c->string, tar_c->c_color
);
#endif

	    }

	    free(c_string);
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}


	/* Allocate XPM image data */
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToGreyscale():\
 Generating XpmImage data %ix%i  %i bytes...\n",
 width, height,
 width * height * sizeof(unsigned int)
);
#endif
	img->data = (unsigned int *)malloc(
	    width * height * sizeof(unsigned int)
	);
	if(img->data == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}

	/* Copy/converting the source RGBA image data to the target XPM
	 * color-index image data
	 */
	if(img->data != NULL)
	{
	    u_int32_t transparent_rgba = 0x00000000, grey_rgba, g;
	    unsigned int xpm_transparent_pix = ImgXPMColormapMatchColorFromRGBA(
		rgba_color, total_rgba_colors,
		(const u_int8_t *)&transparent_rgba, 0
	    );
	    unsigned int *tar_ptr = img->data;
	    int y = 0;
	    const u_int8_t      *src_line = rgba,
				*src_line_end = src_line + (bpl * height),
				*src_ptr, *src_end;
	    while(src_line < src_line_end)
	    {
		src_ptr = src_line;
		src_end = src_ptr + (width * bpp);
		while(src_ptr < src_end)
		{
		    if(src_ptr[3] < threshold)
		    {
		        *tar_ptr++ = xpm_transparent_pix;
		    }
		    else
		    {
			g = ((u_int32_t)src_ptr[0] + (u_int32_t)src_ptr[1] +
			    (u_int32_t)src_ptr[2]) / 3;
			grey_rgba = (u_int32_t)(
			    0xff000000 |
			    (g << 16) | (g << 8) | (g << 0)
			);
		        *tar_ptr++ = ImgXPMColormapMatchColorFromRGBA(
			    rgba_color, total_rgba_colors,
			    (const u_int8_t *)&grey_rgba,
			    0		/* Closeness, 0 for exact */
			);
		    }
		    src_ptr += bpp;
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			total_rgba_colors + y,
			total_rgba_colors + height, 
			width, height,
			bpl, bpp,
			rgba
		    ))
		    {
			*user_aborted = 1;
			break;
		    }
		}

		y++;
		src_line += bpl;
	    }
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}

	/* Write XpmImage and XpmInfo to the XPM file */
	if(ImgXPMWriteFileXPMImage(filename, c_id, img, info))
	{
	    FREE_ALL
	    return(-1);
	}
	else
	{
	    FREE_ALL
	    return(0);
	}
#undef FREE_ALL
}

/*
 *	Writes the specified image data to an XPM file in black & white.
 */
static int ImgXPMWriteFileRGBAToBW(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments,
	int max_colors,		/* -1 for no color limit */
	u_int8_t threshold,
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
)
{
	unsigned int total_rgba_colors = 0;
	rgba_color_struct **rgba_color = NULL;
	XpmInfo *info = NULL;
	XpmImage *img = NULL;

#define FREE_ALL	{		\
 unsigned int i;			\
					\
 /* Delete XPM image */			\
 if(img != NULL) {			\
  XpmFreeXpmImage(img);			\
  free(img);				\
  img = NULL;				\
 }					\
					\
 /* Delete XPM info */			\
 if(info != NULL) {			\
  XpmFreeXpmInfo(info);			\
  free(info);				\
  info = NULL;				\
 }					\
					\
 /* Delete RGBA colormap */		\
 for(i = 0; i < total_rgba_colors; i++) { \
  free(rgba_color[i]);			\
 }					\
 free(rgba_color);			\
 rgba_color = NULL;			\
 total_rgba_colors = 0;			\
					\
}

	/* Begin generating colormap */

	/* Generate symbolic colors */
	if(1)
	{
	    u_int8_t pix[4] = { 0x00, 0x00, 0x00, 0x00 };
	    ImgXPMColormapColorAppend(   
		&rgba_color, &total_rgba_colors,
		pix
	    );
	}

	/* Generate black & white RGBA colormap from the source
	 * RGBA image
	 */
	if(1)
	{
	    int y = 0;
	    u_int32_t g, bw_rgba;
	    const u_int8_t	*line = rgba,
				*line_end = line + (bpl * height),
				*ptr, *ptr_end;

	    /* Iterate through the source RGBA image and generate the
	     * RGBA colormap
	     */
	    while(line < line_end)
	    {
		for(ptr = line,
		    ptr_end = ptr + (width * bpp);
		    ptr < ptr_end;
		    ptr += bpp
		)
		{
		    /* Do not generate colors from a source pixel who's
		     * alpha value is below the threshold since those
		     * pixels will always use the transparent color
		     */
		    if(ptr[3] < threshold)
			continue;

		    /* Get grey value of the current RGBA pixel */
		    g = ((u_int32_t)ptr[0] + (u_int32_t)ptr[1] +
			(u_int32_t)ptr[2]) / 3;

		    /* Threshold grey value to black or white */
		    g = (g >= 0x80) ? 0xff : 0x00;

		    /* Generate bw pixel value */
		    bw_rgba = (u_int32_t)(
			0xff000000 |
			(g << 16) | (g << 8) | (g << 0)
		    );

		    /* Is this bw pixel value not in the RGBA
		     * colormap yet?
		     */
		    if(!ImgXPMColormapColorExists(
			rgba_color, total_rgba_colors,
			(const u_int8_t *)&bw_rgba
		    ))
		    { 
			if(ImgXPMColormapColorAppend(
			    &rgba_color, &total_rgba_colors,
			    (const u_int8_t *)&bw_rgba
			) == NULL)
			{
			    FREE_ALL
			    imgio_last_write_error = "Memory allocation error";
			    return(-3);
			}
		    }
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			y,
			height,
			width, height,
			bpl, bpp,
			rgba
		    )) 
		    {
			*user_aborted = 1;
			break;
		    }
		}

		y++;
		line += bpl;
	    }
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);  
	}

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToBW(): Generated colormap with %i colors.\n",
 total_rgba_colors
);
#endif


	/* Create XpmInfo */
	info = (XpmInfo *)calloc(1, sizeof(XpmInfo));
	if(info == NULL)
	{
	    FREE_ALL  
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}
	info->valuemask = XpmComments | XpmHotspot;
	info->hints_cmt = NULL;
	info->colors_cmt = NULL;
	info->pixels_cmt = STRDUP(comments);
	info->x_hotspot = x;
	info->y_hotspot = y;


	/* Create XpmImage */
	img = (XpmImage *)calloc(1, sizeof(XpmImage));
	if(img == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);  
	}
	img->width = width;
	img->height = height;
	if(total_rgba_colors > 0)
	{
	    unsigned int m = XPM_ASCII_CHAR_RANGE;
	    for(img->cpp = 1;
		m < total_rgba_colors;
		img->cpp++, m *= XPM_ASCII_CHAR_RANGE
	    );
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToColor():\
 Generating XPM colormap: ncolors: %i  cpp: %i (%i possible colors)\n",
 total_rgba_colors, img->cpp, m
);
#endif
	}

	/* Create XPM colormap on the XpmImage */
	if((img->cpp > 0) && (total_rgba_colors > 0))
	{
	    const unsigned int cpp = img->cpp;
	    unsigned int i;
	    char *s, *c_string = malloc((cpp + 1) * sizeof(char));
	    const rgba_color_struct *src_c;
	    XpmColor *tar_c;  

	    /* Reset color string */
	    memset(c_string, XPM_ASCII_CHAR_MIN, cpp);
	    c_string[cpp] = '\0';

	    /* Allocate XpmColors on the XPM colormap */
	    img->ncolors = total_rgba_colors;
	    img->colorTable = (XpmColor *)calloc(
		img->ncolors, sizeof(XpmColor)
	    );
	    if(img->colorTable == NULL)
	    {
		free(c_string);
		img->ncolors = 0;
		FREE_ALL
		imgio_last_write_error = "Memory allocation error";
		return(-3);
	    }

	    /* Iterate through RGBA colormap and create XpmColors */
	    for(i = 0; i < total_rgba_colors; i++)
	    {
		src_c = rgba_color[i];
		tar_c = &img->colorTable[i];

		/* Set XpmColor's ASCII characters string that represent
		 * this color
		 */
		tar_c->string = STRDUP(c_string);

		/* Set XpmColor's color value string */
/* TODO check for symbolic colors */
		if(src_c->a == 0x00)
		{
		    tar_c->c_color = STRDUP("None");
		}
		else
		{
		    tar_c->c_color = s = (char *)malloc((7 + 1) * sizeof(char));
		    sprintf(
			s, "#%.2X%.2X%.2X",
			src_c->r, src_c->g, src_c->b
		    );
		}

		/* Increment characters in the color string */
		for(s = c_string; *s != '\0'; s++)
		{
		    if(*s < XPM_ASCII_CHAR_MAX)
		    {
			*s = (*s) + 1;
			break;
		    }
		    else
		    {
			*s = XPM_ASCII_CHAR_MIN;
		    }
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			i,
			total_rgba_colors + height, 
			width, height,
			bpl, bpp,
			rgba
		    ))
		    {
			*user_aborted = 1;
			break;
		    }
		}

#if (DEBUG_LEVEL >= 3)
printf(
 "ImgXPMWriteFileRGBAToBW():\
 Generated color %i: \"%s\"=\"%s\"\n",
 i, tar_c->string, tar_c->c_color
);
#endif

	    }

	    free(c_string);
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}


	/* Allocate XPM image data */ 
#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBAToBW():\
 Generating XpmImage data %ix%i  %i bytes...\n",
 width, height,
 width * height * sizeof(unsigned int)
);
#endif
	img->data = (unsigned int *)malloc(
	    width * height * sizeof(unsigned int)
	);
	if(img->data == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}
	 
	/* Copy/convert the source RGBA image data to the target XPM
	 * color-index image data
	 */
	if(img->data != NULL)
	{
	    u_int32_t transparent_rgba = 0x00000000, g, bw_rgba;
	    unsigned int xpm_transparent_pix = ImgXPMColormapMatchColorFromRGBA(
		rgba_color, total_rgba_colors,
		(const u_int8_t *)&transparent_rgba, 0
	    );
	    unsigned int *tar_ptr = img->data;
	    int y = 0;
	    const u_int8_t      *src_line = rgba,
				*src_line_end = src_line + (bpl * height),
				*src_ptr, *src_end;                       
	    while(src_line < src_line_end)
	    {
		src_ptr = src_line;
		src_end = src_ptr + (width * bpp);
		while(src_ptr < src_end)
		{
		    if(src_ptr[3] < threshold)
		    {
			*tar_ptr++ = xpm_transparent_pix;
                    }
                    else
                    {
                        g = ((u_int32_t)src_ptr[0] + (u_int32_t)src_ptr[1] +
                            (u_int32_t)src_ptr[2]) / 3;
			g = (g >= 0x80) ? 0xff : 0x00;
                        bw_rgba = (u_int32_t)(
                            0xff000000 |
                            (g << 16) | (g << 8) | (g << 0)
                        );
			*tar_ptr++ = ImgXPMColormapMatchColorFromRGBA(
			    rgba_color, total_rgba_colors,
			    (const u_int8_t *)&bw_rgba,
			    0		/* Closeness, 0 for exact */
		        );
		    }
		    src_ptr += bpp;
		}

		/* Report progress */
		if((progress_cb != NULL) && !(*user_aborted))
		{
		    if(!progress_cb(
			client_data,
			total_rgba_colors + y,
			total_rgba_colors + height,
			width, height,
			bpl, bpp,
			rgba
		    ))
		    { 
			*user_aborted = 1;
			break;
		    }
		}

		y++;
		src_line += bpl;
	    }
	}

	/* User aborted? */
	if(*user_aborted)
	{
	    FREE_ALL
	    return(-4);
	}

	/* Write XpmImage and XpmInfo to the XPM file */
	if(ImgXPMWriteFileXPMImage(filename, c_id, img, info))
	{
	    FREE_ALL
	    return(-1);
	}
	else
	{
	    FREE_ALL
	    return(0);
	}
#undef FREE_ALL
}

/*
 *	Writes the specified image data to an XPM file.
 *
 *	Return values:
 *
 *      0       Success
 *      -1      General error
 *      -2      Bad value
 *      -3      Systems error
 *      -4      User abort
 */
int ImgXPMWriteFileRGBA(
	const char *filename,
	int width, int height,
	int bpl, int bpp,
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes RGBA */
	int x, int y, int base_width, int base_height,
	const char *c_id, const char *comments,
	int format,		/* 0 = Black & White
				 * 1 = Greyscale
				 * 2 = Color */
	int max_colors,		/* -1 for no color limit */
	u_int8_t threshold,
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	)
)
{
	int status, user_aborted = 0;
	char *s, *new_c_id;

	/* Reset global last write error message pointer */
	imgio_last_write_error = NULL;

	if((filename == NULL) || (rgba == NULL) ||
	   (width <= 0) || (height <= 0) ||
	   (bpp != 4)
	)
	{
	    imgio_last_write_error = "Invalid value used to describe image";
	    return(-1); 
	}

	/* Automatically calculate bytes per line? */
	if(bpl <= 0)   
	    bpl = width * bpp;

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBA(): Writing XPM file \"%s\"...\n",
 filename
);
#endif

	/* Report initial progress */
	if(progress_cb != NULL)
	{
	    if(!progress_cb(
		client_data,
		0, height,
		width, height,
		bpl, bpp,
		rgba
	    ))
		user_aborted = 1;
	}                        

	/* Generate C ID as needed */
	if(STRISEMPTY(c_id))
	{
	    const char *s = strrchr(filename, '/');
	    new_c_id = (s != NULL) ? STRDUP(s + 1) : STRDUP(filename);
	}
	else
	{
	    new_c_id = STRDUP(c_id);
	}
	/* Replace unacceptable characters in the C ID */
	for(s = new_c_id; *s != '\0'; s++)                   
	{
	    if(!isalnum(*s))
		*s = '_';
	}


#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBA(): C ID=\"%s\"\n",
 new_c_id
);
#endif

	/* Save by format */
	status = -2;
	switch(format)
	{
	  case 0:
	    status = ImgXPMWriteFileRGBAToBW(
		filename,
		width, height, bpl, bpp,
		rgba,
		bg_color,
		x, y, base_width, base_height,
		new_c_id, comments,
		max_colors, threshold,
		client_data, progress_cb, &user_aborted
	    );
	    break;

	  case 1:
	    status = ImgXPMWriteFileRGBAToGreyscale(
		filename,
		width, height, bpl, bpp,
		rgba,
		bg_color,
		x, y, base_width, base_height,
		new_c_id, comments,
		max_colors, threshold,
		client_data, progress_cb, &user_aborted
	    );
	    break;

	  case 2:
	    status = ImgXPMWriteFileRGBAToColor(
		filename,
		width, height, bpl, bpp,
		rgba,
		bg_color,
		x, y, base_width, base_height,
		new_c_id, comments,
		max_colors, threshold,
		client_data, progress_cb, &user_aborted
	    );
	    break;

	  default:
	    imgio_last_write_error = "Unsupported XPM color format";
	    status = -2;
	    break;
	}

	/* Report final progress */
	if((progress_cb != NULL) && !user_aborted)
	{
	    if(!progress_cb(
		client_data,
		height, height,
		width, height,
		bpl, bpp,
		rgba
	    ))
		user_aborted = 1;
	}

	free(new_c_id);

#if (DEBUG_LEVEL >= 1)
printf(
 "ImgXPMWriteFileRGBA(): Done.\n"
);
#endif

	if(user_aborted)
	{
	    imgio_last_write_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(status);
}

#endif	/* HAVE_LIBXPM */
